@@ -38,3 +38,9 @@ SNAPSHOT_SORT_KEY_VALUES = (
     'progress',
     'name', 'display_name',
 )
+
+EXPERIMENTAL_HTTP_HEADER = 'X-OpenStack-Manila-API-Experimental'
+MAX_API_VERSION = '2.5'
+V1_API_VERSION = '1.0'
+V1_SERVICE_TYPE = 'share'
+V2_SERVICE_TYPE = 'sharev2'
