@@ -15,6 +15,7 @@
 
 from __future__ import print_function
 
+
 import os
 import sys
 import time
@@ -22,6 +23,7 @@ import time
 from oslo_utils import strutils
 import six
 
+from manilaclient import api_versions
 from manilaclient.common import constants
 from manilaclient import exceptions
 from manilaclient.openstack.common.apiclient import utils as apiclient_utils
@@ -94,6 +96,17 @@ def _print_share(cs, share):
     cliutils.print_dict(info)
 
 
+def _find_share_instance(cs, instance):
+    """Get a share instance by ID."""
+    return apiclient_utils.find_resource(cs.share_instances, instance)
+
+
+def _print_share_instance(cs, instance):
+    info = instance._info.copy()
+    info.pop('links', None)
+    cliutils.print_dict(info)
+
+
 def _find_share_snapshot(cs, snapshot):
     """Get a snapshot by ID."""
     return apiclient_utils.find_resource(cs.share_snapshots, snapshot)
@@ -149,6 +162,14 @@ def _extract_key_value_options(args, option_name):
     return result_dict
 
 
+def do_api_version(cs, args):
+    """Display the API version information."""
+    columns = ['ID', 'Status', 'Version', 'Min_version']
+    column_labels = ['ID', 'Status', 'Version', 'Minimum Version']
+    versions = cs.services.api_version()
+    cliutils.print_list(versions, columns, field_labels=column_labels)
+
+
 def do_endpoints(cs, args):
     """Discover endpoints that get returned from the authenticate services."""
     catalog = cs.keystone_client.service_catalog.catalog
@@ -278,7 +299,7 @@ def do_quota_defaults(cs, args):
     default=None,
     help='Whether force update the quota even if the already used '
          'and reserved exceeds the new quota.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_quota_update(cs, args):
     """Update the quotas for a tenant/user."""
 
@@ -309,7 +330,7 @@ def do_quota_delete(cs, args):
     'class_name',
     metavar='<class>',
     help='Name of quota class to list the quotas for.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_quota_class_show(cs, args):
     """List the quotas for a quota class."""
 
@@ -352,14 +373,14 @@ def do_quota_class_show(cs, args):
     type=int,
     default=None,
     help='New value for the "share_networks" quota.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_quota_class_update(cs, args):
     """Update the quotas for a quota class."""
 
     _quota_update(cs.quota_classes, args.class_name, args)
 
 
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_absolute_limits(cs, args):
     """Print a list of absolute limits for a user."""
     limits = cs.limits.get().absolute
@@ -367,7 +388,7 @@ def do_absolute_limits(cs, args):
     cliutils.print_list(limits, columns)
 
 
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_rate_limits(cs, args):
     """Print a list of rate limits for a user."""
     limits = cs.limits.get().rate
@@ -426,7 +447,13 @@ def do_rate_limits(cs, args):
     default=False,
     help="Level of visibility for share. Defines whether other tenants are "
          "able to see it or not.")
-@cliutils.service_type('share')
+@cliutils.arg(
+    '--availability-zone', '--availability_zone', '--az',
+    metavar='<availability-zone>',
+    default=None,
+    action='single_alias',
+    help='Availability zone in which share should be created.')
+@cliutils.service_type('sharev2')
 def do_create(cs, args):
     """Creates a new share (NFS, CIFS, GlusterFS or HDFS)."""
 
@@ -442,10 +469,28 @@ def do_create(cs, args):
                              metadata=share_metadata,
                              share_network=share_network,
                              share_type=args.share_type,
-                             is_public=args.public)
+                             is_public=args.public,
+                             availability_zone=args.availability_zone)
     _print_share(cs, share)
 
 
+@cliutils.arg(
+    'share', metavar='<share>', help='Name or ID of share to migrate.')
+@cliutils.arg('host', metavar='<host#pool>', help='Destination host and pool.')
+@cliutils.arg('--force-host-copy', metavar='<True|False>',
+              choices=['True', 'False'], required=False,
+              help='Enables or disables generic host-based '
+              'force-migration, which bypasses driver '
+              'optimizations. Default=False.',
+              default=False)
+@cliutils.service_type('sharev2')
+@api_versions.experimental_api
+def do_migrate(cs, args):
+    """Migrates share to a new host."""
+    share = _find_share(cs, args.share)
+    share.migrate_share(args.host, args.force_host_copy)
+
+
 @cliutils.arg(
     'share',
     metavar='<share>',
@@ -461,7 +506,7 @@ def do_create(cs, args):
     nargs='+',
     default=[],
     help='Metadata to set or unset (key is only necessary on unset).')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_metadata(cs, args):
     """Set or delete metadata on a share."""
     share = _find_share(cs, args.share)
@@ -477,12 +522,12 @@ def do_metadata(cs, args):
     'share',
     metavar='<share>',
     help='Name or ID of the share.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_metadata_show(cs, args):
     """Show metadata of given share."""
     share = _find_share(cs, args.share)
     metadata = cs.shares.get_metadata(share)._info
-    cliutils.print_dict(metadata, 'Metadata-property')
+    cliutils.print_dict(metadata, 'Property')
 
 
 @cliutils.arg(
@@ -495,20 +540,20 @@ def do_metadata_show(cs, args):
     nargs='+',
     default=[],
     help='Metadata entry or entries to update.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_metadata_update_all(cs, args):
     """Update all metadata of a share."""
     share = _find_share(cs, args.share)
     metadata = _extract_metadata(args)
     metadata = share.update_all_metadata(metadata)._info['metadata']
-    cliutils.print_dict(metadata, 'Metadata-property')
+    cliutils.print_dict(metadata, 'Property')
 
 
 @cliutils.arg(
     'service_host',
     metavar='<service_host>',
     type=str,
-    help='manage-share service host: some.host@driver[#pool]')
+    help='manage-share service host: some.host@driver#pool')
 @cliutils.arg(
     'protocol',
     metavar='<protocol>',
@@ -518,7 +563,8 @@ def do_metadata_update_all(cs, args):
     'export_path',
     metavar='<export_path>',
     type=str,
-    help='Share export path.')
+    help='Share export path, NFS share such as: 10.0.0.1:/foo_path, '
+         'CIFS share such as: \\10.0.0.1\foo_name_of_cifs_share')
 @cliutils.arg(
     '--name',
     metavar='<name>',
@@ -531,7 +577,7 @@ def do_metadata_update_all(cs, args):
     default=None)
 @cliutils.arg(
     '--share_type', '--share-type',
-    metavar='<share_type>',
+    metavar='<share-type>',
     default=None,
     action='single_alias',
     help='Optional share type assigned to share. (Default=None)')
@@ -613,7 +659,7 @@ def do_force_delete(cs, args):
     'share',
     metavar='<share>',
     help='Name or ID of the NAS share.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_show(cs, args):
     """Show details about a NAS share."""
     share = _find_share(cs, args.share)
@@ -642,7 +688,7 @@ def do_show(cs, args):
     choices=['rw', 'ro'],
     help='Share access level ("rw" and "ro" access levels are supported). '
          'Defaults to None.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_access_allow(cs, args):
     """Allow access to the share."""
     share = _find_share(cs, args.share)
@@ -658,7 +704,7 @@ def do_access_allow(cs, args):
     'id',
     metavar='<id>',
     help='ID of the access rule to be deleted.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_access_deny(cs, args):
     """Deny access to a share."""
     share = _find_share(cs, args.share)
@@ -669,7 +715,7 @@ def do_access_deny(cs, args):
     'share',
     metavar='<share>',
     help='Name or ID of the share.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_access_list(cs, args):
     """Show access list for share."""
     share = _find_share(cs, args.share)
@@ -800,12 +846,12 @@ def do_access_list(cs, args):
     action='store_true',
     default=False,
     help="Add public shares from all tenants to result.")
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_list(cs, args):
     """List NAS shares with filters."""
     list_of_keys = [
         'ID', 'Name', 'Size', 'Share Proto', 'Status', 'Is Public',
-        'Share Type', 'Export location', 'Host',
+        'Share Type', 'Export location', 'Host', 'Availability Zone'
     ]
     all_tenants = int(os.environ.get("ALL_TENANTS", args.all_tenants))
 
@@ -842,6 +888,79 @@ def do_list(cs, args):
     cliutils.print_list(shares, list_of_keys)
 
 
+@cliutils.arg(
+    '--share-id',
+    '--share_id',  # alias
+    metavar='<share_id>',
+    default=None,
+    action='single_alias',
+    help='Filter results by share ID.')
+@cliutils.service_type('sharev2')
+def do_share_instance_list(cs, args):
+    """List share instances."""
+    share = _find_share(cs, args.share_id) if args.share_id else None
+
+    list_of_keys = [
+        'ID', 'Share ID', 'Host', 'Status', 'Availability Zone',
+        'Share Network ID', 'Share Server ID'
+    ]
+    if share:
+        instances = cs.shares.list_instances(share)
+    else:
+        instances = cs.share_instances.list()
+
+    cliutils.print_list(instances, list_of_keys)
+
+
+@cliutils.arg(
+    'instance',
+    metavar='<instance>',
+    help='Name or ID of the share instance.')
+@cliutils.service_type('sharev2')
+def do_share_instance_show(cs, args):
+    """Show details about a share instance."""
+    instance = _find_share_instance(cs, args.instance)
+    _print_share_instance(cs, instance)
+
+
+@cliutils.arg(
+    'instance',
+    metavar='<instance>',
+    nargs='+',
+    help='Name or ID of the instance(s) to force delete.')
+def do_share_instance_force_delete(cs, args):
+    """Attempt force-delete of share instance, regardless of state."""
+    failure_count = 0
+    for instance in args.instance:
+        try:
+            _find_share_instance(cs, instance).force_delete()
+        except Exception as e:
+            failure_count += 1
+            print("Delete for share instance %s failed: %s" % (instance, e),
+                  file=sys.stderr)
+    if failure_count == len(args.instance):
+        raise exceptions.CommandError("Unable to force delete any of "
+                                      "specified share instances.")
+
+
+@cliutils.arg(
+    'instance',
+    metavar='<instance>',
+    help='Name or ID of the share instance to modify.')
+@cliutils.arg(
+    '--state',
+    metavar='<state>',
+    default='available',
+    help=('Indicate which state to assign the instance. Options include '
+          'available, error, creating, deleting, error_deleting. If no '
+          'state is provided, available will be used.'))
+@cliutils.service_type('sharev2')
+def do_share_instance_reset_state(cs, args):
+    """Explicitly update the state of a share instance."""
+    instance = _find_share_instance(cs, args.instance)
+    instance.reset_state(args.state)
+
+
 @cliutils.arg(
     '--all-tenants',
     dest='all_tenants',
@@ -911,7 +1030,7 @@ def do_list(cs, args):
     action='single_alias',
     help='Sort direction, available values are %(values)s. '
          'OPTIONAL: Default=None.' % {'values': constants.SORT_DIR_VALUES})
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_snapshot_list(cs, args):
     """List all the snapshots."""
     list_of_keys = [
@@ -941,7 +1060,7 @@ def do_snapshot_list(cs, args):
     'snapshot',
     metavar='<snapshot>',
     help='Name or ID of the snapshot.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_snapshot_show(cs, args):
     """Show details about a snapshot."""
     snapshot = _find_share_snapshot(cs, args.snapshot)
@@ -969,7 +1088,7 @@ def do_snapshot_show(cs, args):
     metavar='<description>',
     default=None,
     help='Optional snapshot description. (Default=None)')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_snapshot_create(cs, args):
     """Add a new snapshot."""
     share = _find_share(cs, args.share)
@@ -1002,7 +1121,7 @@ def do_snapshot_create(cs, args):
     type=str,
     action="single_alias",
     help='Public share is visible for all tenants.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_update(cs, args):
     """Rename a share."""
     kwargs = {}
@@ -1034,7 +1153,7 @@ def do_update(cs, args):
     metavar='<description>',
     help='Optional snapshot description. (Default=None)',
     default=None)
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_snapshot_rename(cs, args):
     """Rename a snapshot."""
     kwargs = {}
@@ -1053,7 +1172,7 @@ def do_snapshot_rename(cs, args):
     'snapshot',
     metavar='<snapshot>',
     help='Name or ID of the snapshot to delete.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_snapshot_delete(cs, args):
     """Remove a snapshot."""
     snapshot = _find_share_snapshot(cs, args.snapshot)
@@ -1064,7 +1183,7 @@ def do_snapshot_delete(cs, args):
     'snapshot',
     metavar='<snapshot>',
     help='Name or ID of the snapshot to force delete.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_snapshot_force_delete(cs, args):
     """Attempt force-delete of snapshot, regardless of state."""
     snapshot = _find_share_snapshot(cs, args.snapshot)
@@ -1083,7 +1202,7 @@ def do_snapshot_force_delete(cs, args):
           'Options include available, error, creating, deleting, '
           'error_deleting. If no state is provided, '
           'available will be used.'))
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_snapshot_reset_state(cs, args):
     """Explicitly update the state of a snapshot."""
     snapshot = _find_share_snapshot(cs, args.snapshot)
@@ -1101,7 +1220,7 @@ def do_snapshot_reset_state(cs, args):
     help=('Indicate which state to assign the share. Options include '
           'available, error, creating, deleting, error_deleting. If no '
           'state is provided, available will be used.'))
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_reset_state(cs, args):
     """Explicitly update the state of a share."""
     share = _find_share(cs, args.share)
@@ -1768,6 +1887,13 @@ def _print_type_required_extra_specs(share_type):
         return "N/A"
 
 
+def _print_type_optional_extra_specs(share_type):
+    try:
+        return _print_dict(share_type.get_optional_keys())
+    except exceptions.NotFound:
+        return "N/A"
+
+
 def _print_share_type_list(stypes, default_share_type=None):
 
     def _is_default(share_type):
@@ -1783,13 +1909,21 @@ def _print_share_type_list(stypes, default_share_type=None):
         'Visibility': is_public,
         'is_default': _is_default,
         'required_extra_specs': _print_type_required_extra_specs,
+        'optional_extra_specs': _print_type_optional_extra_specs,
     }
 
     for stype in stypes:
         stype = stype.to_dict()
         stype['Visibility'] = stype.pop('is_public', 'unknown')
 
-    fields = ['ID', 'Name', 'Visibility', 'is_default', 'required_extra_specs']
+    fields = [
+        'ID',
+        'Name',
+        'Visibility',
+        'is_default',
+        'required_extra_specs',
+        'optional_extra_specs',
+    ]
     cliutils.print_list(stypes, fields, formatters)
 
 
@@ -1806,7 +1940,7 @@ def _find_share_type(cs, stype):
     return apiclient_utils.find_resource(cs.share_types, stype)
 
 
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 @cliutils.arg(
     '--all',
     dest='all',
@@ -1824,7 +1958,7 @@ def do_type_list(cs, args):
     _print_share_type_list(stypes, default_share_type=default)
 
 
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_extra_specs_list(cs, args):
     """Print a list of current 'share types and extra specs' (Admin Only)."""
     stypes = cs.share_types.list()
@@ -1838,30 +1972,47 @@ def do_extra_specs_list(cs, args):
 @cliutils.arg(
     'spec_driver_handles_share_servers',
     metavar='<spec_driver_handles_share_servers>',
-    default='',
-    nargs='?',
+    type=str,
     help="Required extra specification. "
-         "Valid values 'true'/'1' and 'false'/'0'")
+         "Valid values are 'true'/'1' and 'false'/'0'")
+@cliutils.arg(
+    '--snapshot_support',
+    '--snapshot-support',
+    metavar='<snapshot_support>',
+    action='single_alias',
+    help="Boolean extra spec that used for filtering of back ends by their "
+         "capability to create share snapshots. (Default is True).")
 @cliutils.arg(
     '--is_public',
     '--is-public',
     metavar='<is_public>',
     action='single_alias',
     help="Make type accessible to the public (default true).")
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_type_create(cs, args):
     """Create a new share type."""
-
+    kwargs = {
+        "name": args.name,
+        "is_public": strutils.bool_from_string(args.is_public, default=True),
+    }
     try:
-        extra_spec = strutils.bool_from_string(
-            args.spec_driver_handles_share_servers, strict=True)
+        kwargs['spec_driver_handles_share_servers'] = (
+            strutils.bool_from_string(
+                args.spec_driver_handles_share_servers, strict=True))
     except ValueError as e:
         msg = ("Argument spec_driver_handles_share_servers "
                "argument is not valid: %s" % six.text_type(e))
         raise exceptions.CommandError(msg)
+    try:
+        if args.snapshot_support:
+            kwargs['spec_snapshot_support'] = strutils.bool_from_string(
+                args.snapshot_support, strict=True)
+    except ValueError as e:
+        msg = ("Argument 'snapshot_support' is of boolean type and has "
+               "invalid value: %s" % six.text_type(e))
+        raise exceptions.CommandError(msg)
 
-    is_public = strutils.bool_from_string(args.is_public, default=True)
-    stype = cs.share_types.create(args.name, extra_spec, is_public=is_public)
+    stype = cs.share_types.create(**kwargs)
     _print_share_type_list([stype])
 
 
@@ -1869,7 +2020,7 @@ def do_type_create(cs, args):
     'id',
     metavar='<id>',
     help="Name or ID of the share type to delete.")
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_type_delete(cs, args):
     """Delete a specific share type."""
     share_type = _find_share_type(cs, args.id)
@@ -1891,7 +2042,7 @@ def do_type_delete(cs, args):
     nargs='*',
     default=None,
     help='Extra_specs to set or unset (key is only necessary on unset).')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_type_key(cs, args):
     """Set or unset extra_spec for a share type."""
     stype = _find_share_type(cs, args.stype)
@@ -1923,7 +2074,7 @@ def do_type_key(cs, args):
     type=str,
     default='.*',
     help='Filter results by pool name.  Regular expressions are supported.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_pool_list(cs, args):
     """List all backend storage pools known to the scheduler (Admin only)."""
 
@@ -1941,7 +2092,7 @@ def do_pool_list(cs, args):
     'share_type',
     metavar='<share_type>',
     help="Filter results by share type name or ID.")
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_type_access_list(cs, args):
     """Print access information about the given share type."""
     share_type = _find_share_type(cs, args.share_type)
@@ -1963,7 +2114,7 @@ def do_type_access_list(cs, args):
     'project_id',
     metavar='<project_id>',
     help='Project ID to add share type access for.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_type_access_add(cs, args):
     """Adds share type access for the given project."""
     vtype = _find_share_type(cs, args.share_type)
@@ -1979,7 +2130,7 @@ def do_type_access_add(cs, args):
     'project_id',
     metavar='<project_id>',
     help='Project ID to remove share type access for.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_type_access_remove(cs, args):
     """Removes share type access for the given project."""
     vtype = _find_share_type(cs, args.share_type)
@@ -1993,8 +2144,21 @@ def do_type_access_remove(cs, args):
               metavar='<new_size>',
               type=int,
               help='New size of share, in GBs.')
-@cliutils.service_type('share')
+@cliutils.service_type('sharev2')
 def do_extend(cs, args):
     """Increases the size of an existing share."""
     share = _find_share(cs, args.share)
     cs.shares.extend(share, args.new_size)
+
+
+@cliutils.arg('share', metavar='<share>',
+              help='Name or ID of share to shrink.')
+@cliutils.arg('new_size',
+              metavar='<new_size>',
+              type=int,
+              help='New size of share, in GBs.')
+@cliutils.service_type('sharev2')
+def do_shrink(cs, args):
+    """Decreases the size of an existing share."""
+    share = _find_share(cs, args.share)
+    cs.shares.shrink(share, args.new_size)
